from datetime import datetime, timedelta, timezone
from logging import getLogger, DEBUG

from django.conf import settings
from django.db.models import F, Subquery

from pulpcore.plugin.models import Content, ProgressReport
from pulpcore.plugin.constants import TASK_STATES
from pulpcore.plugin.models import (
    GroupProgressReport,
    RepositoryContent,
    TaskGroup,
)
from pulpcore.plugin.tasking import dispatch
from pulp_rpm.app.models.package import Package
from pulp_rpm.app.models.repository import RpmRepository
from pulp_rpm.app.shared_utils import annotate_with_age

log = getLogger(__name__)


def prune_repo_packages(repo_pk, keep_days, dry_run):
    """
    This task prunes old Packages from the latest_version of the specified repository.

    Args:
        repo_pk (UUID): UUID of the RpmRepository to be pruned.
        keep_days(int): Keep RepositoryContent created less than this many days ago.
        dry_run (boolean): If True, don't actually do the prune, just log to-be-pruned Packages.
    """
    repo = RpmRepository.objects.filter(pk=repo_pk).get()
    curr_vers = repo.latest_version()
    eldest_datetime = datetime.now(tz=timezone.utc) - timedelta(days=keep_days)
    log.info(f"PRUNING REPOSITORY {repo.name}.")
    log.debug(f">>> TOTAL RPMS: {curr_vers.get_content(Package.objects).count()}")

    # We only care about RPM-Names that have more than one EVRA - "singles" are always kept.
    rpm_by_name_age = (
        annotate_with_age(curr_vers.get_content(Package.objects))
        .filter(age__gt=1)
        .order_by("name", "epoch", "version", "release", "arch")
        .values("pk")
    )
    log.debug(f">>> NAME/AGE COUNT {rpm_by_name_age.count()}")
    log.debug(
        ">>> # NAME/ARCH w/ MULTIPLE EVRs: {}".format(
            curr_vers.get_content(Package.objects)
            .filter(pk__in=rpm_by_name_age)
            .values("name", "arch")
            .distinct()
            .count()
        )
    )
    log.debug(
        ">>> # UNIQUE NAME/ARCHS: {}".format(
            curr_vers.get_content(Package.objects).values("name", "arch").distinct().count()
        )
    )

    # Find the RepositoryContents associated with the multi-EVR-names from above,
    # whose maximum-pulp-created date is LESS THAN eldest_datetime.
    #
    # Note that we can "assume" the latest-date is an "add" with no "remove", since we're
    # limiting ourselves to the list of ids that we know are in the repo's current latest-version!
    target_ids_q = (
        RepositoryContent.objects.filter(
            content__in=Subquery(rpm_by_name_age), repository=repo, version_removed=None
        )
        .filter(pulp_created__lt=eldest_datetime)
        .values_list("content_id", flat=True)
    )
    to_be_removed = len(target_ids_q)
    log.debug(f">>> TARGET IDS: {to_be_removed}.")
    # Use the progressreport to report back numbers. The prune happens as one
    # action.
    data = dict(
        message=f"Pruning {repo.name}",
        code="rpm.package.prune.repository",
        total=to_be_removed,
        state=TASK_STATES.COMPLETED,
        done=0,
    )

    if dry_run:
        if log.getEffectiveLevel() == DEBUG:  # Don't go through the loop unless debugging
            log.debug(">>> Packages to be removed : ")
            for p in (
                Package.objects.filter(pk__in=target_ids_q)
                .order_by("name", "epoch", "version", "release", "arch")
                .values("name", "epoch", "version", "release", "arch")
            ):
                log.debug(f'{p["name"]}-{p["epoch"]}:{p["version"]}-{p["release"]}.{p["arch"]}')
    else:
        with repo.new_version(base_version=None) as new_version:
            new_version.remove_content(Content.objects.filter(pk__in=target_ids_q))
        data["done"] = to_be_removed

    pb = ProgressReport(**data)
    pb.save()

    # Report back that this repo has completed.
    gpr = TaskGroup.current().group_progress_reports.filter(code="rpm.package.prune")
    gpr.update(done=F("done") + 1)


def prune_packages(
    repo_pks,
    keep_days=14,
    dry_run=False,
):
    """
    This task prunes old Packages from the latest_version of the specified list of repos.

    "Old" in this context is defined by the RepositoryContent record that added a Package
    to the repository in question.

    It will issue one task-per-repository.

    Kwargs:
        repo_pks (list): A list of repo pks the pruning is performed on.
        keep_days(int): Keep RepositoryContent created less than this many days ago.
        repo_concurrency (int): number of repos to prune at a time.
        dry_run (boolean): If True, don't actually do the prune, just record to-be-pruned Packages..
    """

    repos_to_prune = RpmRepository.objects.filter(pk__in=repo_pks)
    task_group = TaskGroup.current()

    # We want to be able to limit the number of available-workers that prune will consume,
    # so that pulp can continue to work while pruning many repositories. We accomplish this by
    # creating a reserved-resource string for each repo-prune-task based on that repo's index in
    # the dispatch loop, mod number-of-workers-to-consume.
    #
    # By default, prune will consume up to 5 workers.
    #
    # (This comment and code below based on
    #   https://github.com/pulp/pulpcore/blob/main/pulpcore/app/tasks/importer.py#L503-L512
    # When we have a generic-approach to throttling mass-task-spawning, both places should
    # be refactored to take advantage thereof.
    prune_workers = int(settings.get("PRUNE_WORKERS_MAX", 5))

    gpr = GroupProgressReport(
        message="Pruning old Packages",
        code="rpm.package.prune",
        total=len(repo_pks),
        done=0,
        task_group=task_group,
    )
    gpr.save()

    # Dispatch a task-per-repository.
    # Lock on the the repository *and* to insure the max-concurrency specified.
    # This will keep an "all repositories" prune from locking up all the workers
    # until all repositories are completed.
    for index, a_repo in enumerate(repos_to_prune):
        worker_rsrc = f"rpm-prune-worker-{index % prune_workers}"
        exclusive_resources = [worker_rsrc, a_repo]

        dispatch(
            prune_repo_packages,
            exclusive_resources=exclusive_resources,
            args=(
                a_repo.pk,
                keep_days,
                dry_run,
            ),
            task_group=task_group,
        )



from types import SimpleNamespace

import subprocess

from pulp_rpm.app.tasks.signing import _verify_package_fingerprint


def test_verify_package_fingerprint_accepts_rpm_fingerprint(monkeypatch, tmp_path):
    rpm_output = """\
    Header OpenPGP V4 signature, key fingerprint: c6e7f081cf80e13146676e88829b606631645531: OK
    Header SHA256 digest: OK
    Payload SHA256 digest: OK
    """
    fake_result = SimpleNamespace(stdout=rpm_output, stderr="")

    def fake_run(*args, **kwargs):
        return fake_result

    monkeypatch.setattr(subprocess, "run", fake_run)
    package = SimpleNamespace(name=str(tmp_path / "example.rpm"))
    assert _verify_package_fingerprint(package, "c6e7f081cf80e13146676e88829b606631645531")


def test_verify_package_fingerprint_accepts_long_key_id(monkeypatch, tmp_path):
    rpm_output = """\
    Header OpenPGP V4 RSA/SHA256 signature, key ID 999f7cbf38ab71f4: NOKEY
    Header SHA256 digest: OK
    Payload SHA256 digest: OK
    Legacy OpenPGP V4 RSA/SHA256 signature, key ID 999f7cbf38ab71f4: NOKEY
    """
    fake_result = SimpleNamespace(stdout=rpm_output, stderr="")

    def fake_run(*args, **kwargs):
        return fake_result

    monkeypatch.setattr(subprocess, "run", fake_run)
    package = SimpleNamespace(name=str(tmp_path / "example.rpm"))
    assert _verify_package_fingerprint(package, "999f7cbf38ab71f4")



# Generated by Django 5.2.5 on 2026-01-29 19:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0067_add_rpm_package_signing_result'),
    ]

    operations = [
        migrations.AlterField(
            model_name='rpmpublication',
            name='compression_type',
            field=models.TextField(choices=[('zstd', 'zstd'), ('gz', 'gz'), ('none', 'none')], null=True),
        ),
        migrations.AlterField(
            model_name='rpmrepository',
            name='compression_type',
            field=models.TextField(choices=[('zstd', 'zstd'), ('gz', 'gz'), ('none', 'none')], null=True),
        ),
    ]



import logging
import re
import subprocess
from pathlib import Path
from tempfile import NamedTemporaryFile

from pulpcore.plugin.models import (
    Artifact,
    ContentArtifact,
    CreatedResource,
    PulpTemporaryFile,
    UploadChunk,
    Upload,
)
from pulpcore.plugin.tasking import add_and_remove, general_create
from pulpcore.plugin.util import get_url

from pulp_rpm.app.constants import CHECKSUM_TYPES
from pulp_rpm.app.models.content import RpmPackageSigningResult, RpmPackageSigningService
from pulp_rpm.app.models.package import Package
from pulp_rpm.app.models.repository import RpmRepository


log = logging.getLogger(__name__)


def _save_file(fileobj, final_package):
    with fileobj.file.open() as fd:
        final_package.write(fd.read())
    final_package.flush()


def _save_upload(uploadobj, final_package):
    chunks = UploadChunk.objects.filter(upload=uploadobj).order_by("offset")
    for chunk in chunks:
        final_package.write(chunk.file.read())
        chunk.file.close()
    final_package.flush()


def _verify_package_fingerprint(package_file, signing_fingerprint):
    """Verify if the packge_file is signed with signing_fingerprint or not."""
    completed_process = subprocess.run(
        ("rpm", "-Kv", package_file.name),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
    )
    if completed_process.stderr:
        raise Exception(
            f"Failed to verify package signature: {completed_process.stdout} "
            f"{completed_process.stderr}."
        )

    # check for `key ID` followed by a string of hex digits
    key_ids = re.findall(r"key ID ([0-9A-Fa-f]+)", completed_process.stdout, re.IGNORECASE)
    # check for `key fingerprint:` followed by a string of hex digits
    fingerprints = re.findall(
        r"key fingerprint:\s*([0-9A-Fa-f]+)", completed_process.stdout, re.IGNORECASE
    )
    for candidate in key_ids + fingerprints:
        if signing_fingerprint.lower().endswith(candidate.lower()):
            return True

    return False


def _sign_file(package_file, signing_service, signing_fingerprint):
    result = signing_service.sign(package_file.name, pubkey_fingerprint=signing_fingerprint)
    signed_package_path = Path(result["rpm_package"])
    if not signed_package_path.exists():
        raise Exception(f"Signing script did not create the signed package: {result}")
    artifact = Artifact.init_and_validate(str(signed_package_path))
    artifact.save()
    resource = CreatedResource(content_object=artifact)
    resource.save()
    return artifact


def sign_and_create(
    app_label,
    serializer_name,
    signing_service_pk,
    signing_fingerprint,
    temporary_file_pk,
    *args,
    **kwargs,
):
    data = kwargs.pop("data", None)
    context = kwargs.pop("context", {})

    # Get unsigned package file and sign it
    package_signing_service = RpmPackageSigningService.objects.get(pk=signing_service_pk)
    with NamedTemporaryFile(mode="wb", dir=".", delete=False) as final_package:
        try:
            uploaded_package = PulpTemporaryFile.objects.get(pk=temporary_file_pk)
            _save_file(uploaded_package, final_package)
        except PulpTemporaryFile.DoesNotExist:
            uploaded_package = Upload.objects.get(pk=temporary_file_pk)
            _save_upload(uploaded_package, final_package)

        artifact = _sign_file(final_package, package_signing_service, signing_fingerprint)
    uploaded_package.delete()

    # Create Package content
    data["artifact"] = get_url(artifact)
    # The Package serializer validation method have two branches: the signing and non-signing.
    # Here, the package is already signed, so we need to update the context for a proper validation.
    context["sign_package"] = False
    # The request data is immutable when there's an upload, so we can't delete the upload out of the
    # request data like we do for a file.  Instead, we'll delete it here.
    if "upload" in data:
        del data["upload"]
    general_create(app_label, serializer_name, data=data, context=context, *args, **kwargs)


def signed_add_and_remove(
    repository_pk, add_content_units, remove_content_units, base_version_pk=None
):
    repo = RpmRepository.objects.get(pk=repository_pk)

    if repo.package_signing_service:
        # sign each package and replace it in the add_content_units list
        for package in Package.objects.filter(pk__in=add_content_units).iterator():
            # the viewset is currently already checking (and rejecting) on demand content
            # but in the future we could just download it instead
            content_artifact = package.contentartifact_set.first()
            artifact_obj = content_artifact.artifact
            package_id = str(package.pk)

            with NamedTemporaryFile(mode="wb", dir=".", delete=False) as final_package:
                artifact_file = artifact_obj.file
                _save_file(artifact_file, final_package)

                # check if the package is already signed with our fingerprint
                if _verify_package_fingerprint(final_package, repo.package_signing_fingerprint):
                    continue

                # check if the package has been signed in the past with our fingerprint and replace
                # it with the previously-created signed package if so
                if existing_result := RpmPackageSigningResult.objects.filter(
                    original_package_sha256=content_artifact.artifact.sha256,
                    package_signing_fingerprint=repo.package_signing_fingerprint,
                ).first():
                    while package_id in add_content_units:
                        add_content_units.remove(package_id)
                    add_content_units.append(str(existing_result.result_package.pk))
                    continue

                # create a new signed version of the package
                artifact = _sign_file(
                    final_package, repo.package_signing_service, repo.package_signing_fingerprint
                )
                signed_package = package
                signed_package.pk = None
                signed_package.pulp_id = None
                signed_package.pkgId = artifact.sha256
                signed_package.checksum_type = CHECKSUM_TYPES.SHA256
                signed_package.save()
                ContentArtifact.objects.create(
                    artifact=artifact,
                    content=signed_package,
                    relative_path=content_artifact.relative_path,
                )
                RpmPackageSigningResult.objects.create(
                    original_package_sha256=artifact_obj.sha256,
                    package_signing_fingerprint=repo.package_signing_fingerprint,
                    result_package=signed_package,
                )

                resource = CreatedResource(content_object=signed_package)
                resource.save()
                while package_id in add_content_units:
                    add_content_units.remove(package_id)
                add_content_units.append(str(signed_package.pk))

    return add_and_remove(repository_pk, add_content_units, remove_content_units, base_version_pk)
